from yta_general_utils.downloader import Downloader
from yta_general_utils.image.size import resize_without_scaling
from openai import OpenAI


# TODO: Is this actually useful? I think it could be removed...

def generate_image(prompt, output_filename):
    client = OpenAI()

    response = client.images.generate(
        model = "dall-e-3",
        prompt = prompt,
        size = "1792x1024",
        quality = "standard",
        n = 1,
    )

    image_url = response.data[0].url

    Downloader.download_image(image_url, output_filename)
    resize_without_scaling(output_filename)


"""
    This is one example of use of the alpha transition:

    youtube_url = 'https://www.youtube.com/watch?v=mylQ5Q9GRgs&pp=ygUQYWxwaGEgdHJhbnNpdGlvbg%3D%3D'
    alpha_video = YoutubeVideo(youtube_url).download(output_filename = create_temp_filename('youtube_alpha.mp4'))
    alpha_video = VideoParser.to_moviepy(alpha_video)
    video = VideoParser.to_moviepy('prueba.mp4', False, True).with_subclip(0, 2)
    video2 = VideoParser.to_moviepy('caminando_bosque_60fps.mp4', False, True).with_subclip(0, 2)
    transition = VideoTransition(TransitionMode.PLAYING, duration = 1, method = TransitionMethod.alpha, alpha_video = alpha_video)
    concatenated = TransitionGenerator.apply([video, video2], [transition])
"""
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.video_effect import MoviepySetPosition
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyWith, MoviepyArgument
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_multimedia.video.edition.effect.moviepy.position.objects.video_position import VideoPosition
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.video.frames.numpy_frame_helper import FrameMaskingMethod
from yta_multimedia.video.edition.effect.moviepy.alpha import MoviepyAlphaTransitionHandler
from yta_multimedia.video.edition.duration import set_video_duration, ExtendVideoMode, EnshortVideoMode
from yta_multimedia.video import concatenate_videos
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.math.rate_functions import RateFunction
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.programming.error_message import ErrorMessage
from yta_general_utils.programming.parameter_validator import PythonValidator
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy import ImageClip, CompositeVideoClip, VideoClip
from typing import Callable

import numpy as np


class TransitionMode(Enum):
    """
    Class to represent the transition modes (the way we
    join the videos between a transition).
    """
    FREEZE = 'freeze'
    """
    Freeze the last frame of the first video and the first
    frame ot he second video and make the transition with
    those static frames so the duration is extended by the
    transition duration.
    """
    PLAYING = 'playing'
    """
    Keep both video playing while the transition is run.
    """

class VideoTransition:
    """
    Class to encapsulate the logic to apply in a video 
    transition, containing the mode of transition and 
    the logic to apply on it.
    """
    mode: TransitionMode
    """
    The mode of transition we want to apply.
    """
    duration: float
    """
    The duration that the transition logic will last being
    applied.
    """
    method: Callable
    """
    The method we want to apply as the transition logic, that
    must be a static function of the Transition class.
    """
    kwargs: list
    """
    Other attributes we could need for some specific type of
    transition generation methods.
    """
    def __init__(self, mode: TransitionMode, duration: float, method: Callable, **kwargs):
        """
        The provided 'method' must be one of the static methods
        existing in the Transition class.
        """
        if not PythonValidator.is_class_staticmethod(TransitionMethod, method):
            raise Exception('The provided "method" parameter is not an static method of the Transition class.')
        
        if not NumberValidator.is_positive_number(duration):
            raise Exception(ErrorMessage.parameter_is_not_positive_number('duration'))
        # TODO: Maybe make 'duration' be multiple of frame_time
        # and multiple of a pair frame_time to allow the 
        # transition work equally in both videos

        mode = TransitionMode.to_enum(mode)
        
        self.mode = mode
        self.duration = duration
        self.method = method
        self.kwargs = kwargs

class Transition:
    """
    Class to encapsulate all the transition methods we
    handle in our system. Each transition is a way of
    connecting two different videos.

    This class is built to be used within the
    TransitionGenerator class as parameter to build 
    videos with those transitions.
    """
    video1: VideoClip = None
    """
    The video that comes before the transition.
    """
    transition: VideoTransition = None
    """
    The transition to apply between video1 and video2.
    """
    video2: VideoClip = None
    """
    The video that comes after the transition.
    """
    _final_video: VideoClip = None
    """
    The final video that contains the video1 and video2
    mixed with the transition in the middle. If this
    attribute is set it means that the transition has 
    been applied.
    """
    # TODO: I need to check if I should save the videos
    # and transition clip separately or just the 
    # concatenated one

    @property
    def is_ready(self):
        """
        Return True if the final video has been built by applying
        the desired transition and can be extracted from 'output'
        attribute.
        """
        return self._final_video is not None
    
    @property
    def output(self):
        """
        The output video result of combining the videos and the
        transition in the middle. This will be None if the transition
        has not been applied yet.
        """
        return self._final_video

    def __init__(self, video1: VideoClip, transition: VideoTransition, video2: VideoClip):
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)
        
        if not PythonValidator.is_instance(transition, VideoTransition):
            raise Exception('The provided "transition" is not a VideoTransition instance.')
        
        self.video1 = video1
        self.transition = transition
        self.video2 = video2

    def build(self):
        """
        Build the final video that includes the two videos provided
        with the transition in the middle. This method will build it
        if it is not ready, or return it if it has been built 
        previously.

        The result is a single video that you could use for the next
        transition if needed.
        """
        if not self.is_ready:
            video1, transition, video2 = Transition.create_transition(self.video1, self.video2, self.transition)
            self._final_video = concatenate_videos([video1, transition, video2])

        return self.output

    @staticmethod
    def create_transition(video1, video2, transition: VideoTransition = None):
        """
        Create the transition between 'video1' and 'video2' and 
        return the items in the next order: video1, transition,
        video2. Those items are ready to be concatenated in 
        that order to obtain both videos played with the expected
        transition in between.
        """
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)
        
        if not PythonValidator.is_instance(transition, VideoTransition):
            raise Exception('The provided "transition" is not a VideoTransition instance.')

        # Handle transition and videos
        if transition.mode == TransitionMode.FREEZE:
            # Original videos are not modified
            transition_first_clip = ImageClip(VideoFrameExtractor.get_last_frame(video1), duration = transition.duration).with_fps(60)
            transition_second_clip = ImageClip(VideoFrameExtractor.get_first_frame(video2), duration = transition.duration).with_fps(60)
        elif transition.mode == TransitionMode.PLAYING:
            if transition.duration > video1.duration or transition.duration > video2.duration:
                # TODO: Make this Exception more description (which video is wrong)
                # and talk about that we use the half of the provided duration
                raise Exception(f'The provided "transition.duration" parameter {str(transition.duration)} is not valid according to the provided video duration.')
            
            transition_first_clip = video1.with_subclip(video1.duration - transition.duration, video1.duration)
            transition_second_clip = video2.with_subclip(0, transition.duration)
            video1 = video1.with_subclip(0, video1.duration - transition.duration)
            video2 = video2.with_subclip(transition.duration, video2.duration)
        
        return video1, transition.method(video1 = transition_first_clip, video2 = transition_second_clip, **transition.kwargs), video2
    
class TransitionMethod:
    """
    This class is for internal use only. It is used by
    the TransitionGenerator class to generate clip
    transitions.

    This class encapsulates the functionality to generate 
    transition clips and the different methods to build
    them.
    """
    @staticmethod
    def slide(video1: VideoClip, video2: VideoClip, **kwargs):
        """
        Simple transition in which the last frame of the provided 'video1'
        is replaced by the first frame of the provided 'video2' by sliding
        from right to left.
        """
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)

        # TODO: I was working before in setting the position
        # constants so I could use them to easily handle slides
        # from different sides of the screen

        # TODO: This below is now done in 'to_moviepy' but I 
        # keep the code for the next commit
        # We need to make sure both clips have the same size so,
        # by now, I force it to be in the middle of a black
        # background to fit the other video's size
        # if video1.size > video2.size:
        #     video2 = CompositeVideoClip([
        #         ClipGenerator.get_default_background_video(video1.size, video2.duration, False),
        #         video2.with_position(('center', 'center'))
        #     ])
        # elif video2.size > video1.size:
        #     video1 = CompositeVideoClip([
        #         ClipGenerator.get_default_background_video(video2.size, video1.duration, False),
        #         video1.with_position(('center', 'center'))
        #     ])

        # Transition from last frame of video1 to first of video2
        t1_arg = MoviepyArgument(VideoPosition(0, 0), VideoPosition(-video1.w, 0), TFunctionSetPosition.linear, RateFunction.linear)
        transition_clip_1 = MoviepyWith.apply(video1, with_position = t1_arg)

        t1_position = transition_clip_1.pos(t = 0)
        t2_arg = MoviepyArgument(VideoPosition(video2.w, 0), VideoPosition(t1_position[0], t1_position[1]), TFunctionSetPosition.linear, RateFunction.linear)
        transition_clip_2 = MoviepyWith.apply(video2, with_position = t2_arg)

        return CompositeVideoClip([
            transition_clip_1,
            transition_clip_2
        ])
    
    @staticmethod
    def alpha(video1: VideoClip, video2: VideoClip, alpha_video: VideoClip, alpha_processing_method: FrameMaskingMethod = FrameMaskingMethod.MEAN, **kwargs):
        """
        Transition that involves an alpha clip in between the
        two provided 'video1' and 'video2' clips by applying
        that alpha clip as a mask in the second one to appear
        over the first one.
        """
        video1 = VideoParser.to_moviepy(video1)
        video2 = VideoParser.to_moviepy(video2)
        alpha_video = VideoParser.to_moviepy(alpha_video)
        alpha_processing_method = FrameMaskingMethod.to_enum(alpha_processing_method)

        # We need to adjust the clip to the transition duration
        if (alpha_video.duration != video1.duration):
            alpha_video = set_video_duration(alpha_video, video1.duration, ExtendVideoMode.SLOW_DOWN, EnshortVideoMode.SPEED_UP)

        # TODO: Can we avoid MoviepyAlphaTransitionHandler and use
        # generic classes and .to_mask()' (?)
        mask_clip = MoviepyAlphaTransitionHandler.alpha_clip_to_mask_clip(alpha_video, masking_method = alpha_processing_method)
        video2 = video2.with_mask(mask_clip)

        return CompositeVideoClip([
            video1,
            video2
        ])

class TransitionHandler:
    """
    Class to simplify and encapsulate the functionality
    related to transition between videos. This class has
    been built to simplify the way we apply transitions.
    """
    
    # TODO: Check this below to understand a problem
    """
        There is a big problem here. Even if you put
        the same instance in two different transitions,
        that doesn't mean that you want the second one
        to be updated. You could want them to be different
        transitions using the same video as it is. If you
        want to use the result of a previous Transition you
        should do it by applying the '.build()' result of
        the first one as the input for the next one, and
        you cannot set it that easy

        There is an example below:

        Transition(
            Transition(
                'video',
                VideoTransition(),
                'video2'
            ),
            VideoTransition(),
            'video2'
        )
    """
    @staticmethod
    def apply(videos, transitions: list[VideoTransition] = None):
        """
        Applies the provided 'transitions' between the also provided
        'videos'. The amount of 'transitions' provided must be one 
        less than the 'videos' provided.

        This method will use the final video of each transition as 
        the first video of the next transition as it is placing the
        provided 'transitions' in between all provided 'videos'.

        So, the sequence is the next.
        1. The 'videoX' is played completely.
        2. The transition is played completely.
        3. Go to step 1 with the next video until last one
        4. Last video is played completely
        """
        if not PythonValidator.is_list(videos):
            videos = [VideoParser.to_moviepy(video)]
        
        if len(videos) == 1:
            return videos[0]

        if not PythonValidator.is_list(transitions):
            if not PythonValidator.is_instance(transitions, VideoTransition):
                raise Exception('The provided "transitions" parameter is not a list of VideoTransition nor a single VideoTransition.')
            else:
                transitions = [transitions] * (len(videos) - 1)
        else:
            if len(transitions) != len(videos) - 1:
                raise Exception(f'The number of videos is {str(len(videos))} and the amount of transitions must be {str(len(videos) - 1)} and you provided {str(len(transitions))} transitions.')

        videos = [VideoParser.to_moviepy(video) for video in videos]

        # TODO: Apply the new 'Transition'
        video_transitions = []
        for i in range(1, len(videos)):
            video, video_transition, next_video = Transition.create_transition(videos[i - 1], videos[i], transitions[i - 1])

            videos[i - 1] = video
            videos[i] = next_video
            video_transitions.append(video_transition)
            
        clips_to_concat = []
        for video, video_transition in zip(videos[:-1], video_transitions):
            clips_to_concat.extend([video, video_transition])
        clips_to_concat.append(videos[-1])

        return concatenate_videos(clips_to_concat)


def get_alpha_clip_core_part(alpha_video: VideoClip):
    """
    Process the 'alpha_video' and subclip it to the
    core part only. This part is the one in which
    the animation/effect/transition is happening,
    removing empty frames at the begining or the end
    of the video.

    This method iterates over the frames to detect
    the ones that are not similar to the previous
    ones at the begining and at the end of the 
    provided 'alpha_video'.

    TODO: This method can be updated as it is very
    experimental.
    """
    alpha_video = VideoParser.to_moviepy(alpha_video)

    alpha_frames = list(alpha_video.iter_frames())
    last_frame_index = int(alpha_video.duration * alpha_video.fps) - 1

    previous_frame = None
    first_index = 0
    last_index = last_frame_index

    # Look for non-core part at the begining
    for index, frame in enumerate(alpha_frames):
        if previous_frame is None:
            previous_frame = frame
        else:
            # This calculates the difference
            #print(np.sum(np.abs(frame - previous_frame)))
            if not np.array_equal(frame, previous_frame):
                first_index = index
                break
            else:
                previous_frame = frame

    previous_frame = None
    
    # Look for non-core part at the end
    for index, frame in enumerate(reversed(alpha_frames)):
        if previous_frame is None:
            previous_frame = frame
        else:
            # This calculates the difference
            #print(np.sum(np.abs(frame - previous_frame)))
            if not np.array_equal(frame, previous_frame):
                last_index = last_frame_index - index
                break
            else:
                previous_frame = frame

    alpha_video = alpha_video.with_subclip(first_index * (1 / alpha_video.fps), last_index * (1 / alpha_video.fps))

    # alpha_frames = alpha_frames[first_index:last_index + 1]
    # alpha_video = VideoClip(lambda t: alpha_frames[int(t * alpha_video.fps)], duration = len(alpha_frames) * (1 / alpha_video.fps)).with_fps(alpha_video.fps)

    return alpha_video


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame_extractor import VideoFrameExtractor
from yta_multimedia.alphascreen.masked_clip_creator import MaskedClipCreator
from yta_general_utils.image.region import ImageRegionFinder, Region


# TODO: This class has not been tested yet with real alpha videos
class VideoAlphascreen:
    """
    Class to handle videos with alphascreen regions and insert
    other videos or images on it.
    """
    video = None
    alpha_regions: list[Region] = []

    def __init__(self, video: str):
        video = VideoParser.to_moviepy(video, do_include_mask = True)

        # TODO: Check that video mask has some transparency

        self.video = video
        # TODO: Import from 'frames' the 'get_frame_from_video_as_rgba_by_frame_number(0)'
        self.alpha_regions = ImageRegionFinder.find_transparent_regions(VideoFrameExtractor.get_frame_as_rgba_by_time(video, 0))

        if len(self.alpha_regions) == 0:
            raise Exception('No alpha regions found in the "filename" parameter "{filename}" provided.')
        
        # TODO: What about regions that are just one pixel or too short (?)

        alpha_clip = self.video
        self.masked_clip_creator = MaskedClipCreator(self.alpha_regions, alpha_clip)

    def from_image_to_image(self, image, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'image' in this alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        # TODO: This is not returning RGBA only RGB
        return self.masked_clip_creator.from_image_to_image(image, output_filename)
    
    def from_images_to_image(self, images, output_filename: str = None):
        """
        This method returns a numpy representation of the image
        built by inserting the provided 'images' in this
        alphascreen.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_images_to_image(images, output_filename)
    
    def from_image_to_video(self, image, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'image' fitting the first alphascreen area and centered on
        those areas by applying a mask that let them be seen
        through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_image_to_video(image, duration, output_filename)

    def from_images_to_video(self, images, duration: float, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'images' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_images_to_video(images, duration, output_filename)
    
    def from_video_to_video(self, video, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'video' fitting in the alphascreen area and centered on it
        by applying a mask that let it be seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_video_to_video(video, output_filename)
    
    def from_videos_to_video(self, videos, output_filename: str = None):
        """
        This method returns a CompositeVideoClip with the provided
        'videos' fitting the different alphascreen areas and
        centered on those areas by applying a mask that let them be
        seen through that mask.

        This method will write the result as a local file if the
        'output_filename' parameter is provided.

        This method will work if the amount of images to insert is
        the same of the amount of transparent regions existing in 
        this alphascreen.
        """
        return self.masked_clip_creator.from_videos_to_video(videos, output_filename)


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.filters.frame import FrameFilter
from yta_multimedia.video.parser import VideoParser
from moviepy import Clip
from typing import Union


class BlurEffect(Effect):
    """
    This effect will blur the whole clip. The greater the
    'blur_radius' is, the more blurred it becomes.
    """
    def apply(self, video: Clip, blur_radius: Union[int, None] = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if blur_radius is None:
            blur_radius = 4

        return video.transform(lambda get_frame, t: FrameFilter.blur(get_frame, t, blur_radius = blur_radius))


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.black_and_white_effect import BlackAndWhiteEffect
from yta_multimedia.video.edition.effect.constants import EFFECTS_RESOURCES_FOLDER
from yta_multimedia.resources.video.effect.sound.drive_urls import SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL
from yta_multimedia.resources import Resource
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.programming.parameter_validator import NumberValidator
from moviepy import Clip, ImageClip, AudioFileClip, CompositeVideoClip


class SadMomentEffect(Effect):
    """
    This method gets the first frame of the provided 'clip' and returns a
    new clip that is an incredible 'sad_moment' effect with black and white
    filter, zoom in and rotating effect and also sad violin music.

    The 'duration' parameter is to set the returned clip duration, but the
    default value is a perfect one and it should be keept as it is if 
    possible.

    What to do with the result? **CONCATENATE**

    The result should be added to the clip and must not replace it because
    it freezes one frame so it should continue with the next frame and the
    video as it is.
    """
    # TODO: Is this working (?)
    result_must_replace = False

    def apply(self, video: Clip, duration: float = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlurEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if not duration:
            duration = 4.8

        if not NumberValidator.is_number_between(duration, 0, 120, False):
            raise Exception(f'The provided "duration" parameter "{str(duration)}" is not a valid number between 0 (not included) and 120 (included).')
        
        # We freeze the first frame
        aux = ImageClip(video.get_frame(0), duration = duration)
        aux.fps = video.fps
        video = aux

        # We then build the whole effect
        video = video.with_effects([BlackAndWhiteEffect()]).resized(lambda t: 1 + 0.30 * (t / video.duration)).with_position(lambda t: (-(0.15 * video.w * (t / video.duration)), -(0.15 * video.h * (t / video.duration)))).rotated(lambda t: 5 * (t / video.duration), expand = False)

        # We set the effect audio
        TMP_FILENAME = Resource.get(SAD_MOMENT_GOOGLE_DRIVE_DOWNLOAD_URL, EFFECTS_RESOURCES_FOLDER + 'sounds/sad_moment.mp3')
        video.audio = AudioFileClip(TMP_FILENAME).with_duration(video.duration)

        return CompositeVideoClip([
            ClipGenerator.get_default_background_video(video.size, video.duration),
            video,
        ])


from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.dimensions import adjust_to_aspect_ratio
from yta_general_utils.programming.enum import YTAEnum as Enum
from moviepy.Clip import Clip
from moviepy import CompositeVideoClip
from moviepy.video.fx import Crop
from typing import Union


class EnlargeMode(Enum):
    """
    Enum class to encapsulate the different strategies to
    enlarge a video.
    """
    RESIZE = 'resize'
    """
    The video is resized to fit the expected larger size.
    """
    BACKGROUND = 'background'
    """
    The video is placed as it is but in the middle of a
    black background with the expected size.
    """

class EnshortMode(Enum):
    """
    Enum class to encapsulate the different strategies to
    enshort a video.
    """
    RESIZE = 'resize'
    """
    The video is resized to fit the expected smaller size.
    This can make the video lose the aspect ratio.
    """
    CROP = 'crop'
    """
    The video is cropped in the center to obtain a region
    of the expected size.
    """

def fit_size(video: Clip, size: tuple, enlarge_mode: EnlargeMode = EnlargeMode.RESIZE, enshort_mode: EnshortMode = EnshortMode.RESIZE):
    """
    Make the provided 'video' fit the also provided 'size' by 
    applying the given 'enlarge_mode' or 'enshort_mode' depending
    of the need.
    """
    # TODO: This generates a cyclic import issue
    #video = VideoParser.to_moviepy(video)

    if not enlarge_mode:
        enlarge_mode = EnlargeMode.default()
    else:
        enlarge_mode = EnlargeMode.to_enum(enlarge_mode)

    if not enshort_mode:
        enshort_mode = EnshortMode.default()
    else:
        enshort_mode = EnshortMode.to_enum(enshort_mode)

    if video.size > size:
        # We need to enshort it
        if enshort_mode == EnshortMode.RESIZE:
            video = video.resize(size)
        elif enshort_mode == EnshortMode.CROP:
            # We need to adjust the video to the AR of the given 'size'
            new_ar = size[0] / size[1]
            video_width, video_height = video.size
            new_video_size = adjust_to_aspect_ratio(video_width, video_height, new_ar)

            # TODO: Is this working (?) Changed when moviepy 2.0.0
            video = video.with_effects([Crop(video, width = new_video_size[0], height = new_video_size[1], x_center = video_width / 2, y_center = video_height / 2)])

            video = video.resized((size[0], size[1]))
    elif video.size < size:
        # We need to enlarge it
        if enlarge_mode == EnlargeMode.RESIZE:
            video = video.resize(size)
        elif enlarge_mode == EnlargeMode.BACKGROUND:
            video = CompositeVideoClip([
                ClipGenerator.get_default_background_video(size, video.duration, False),
                video.with_position(('center', 'center'))
            ])

    return video

# TODO: I think I can deprecate (and remove) this method
# when the one above is working and is confirmed
def resize_video(video: Clip, size, output_filename: Union[str, None] = None):
    """
    Resizes the video to the provided 'size' by cropping a
    region of the given 'video' that fits the 'size' aspect
    ratio and resizing that region to the 'size'.

    This method is using the whole video and then resizing,
    so the quality of the video is preserved and no small
    regions are used. The most part of the video is 
    preserved.

    This method returns the video modified.

    This method will write the video if 'output_filename' is
    provided.
    """
    # TODO: This generates a cyclic import issue
    #video = VideoParser.to_moviepy(video)

    # We need to adjust the video to the AR of the given 'size'
    new_ar = size[0] / size[1]
    video_width, video_height = video.size
    new_video_size = adjust_to_aspect_ratio(video_width, video_height, new_ar)

    # TODO: Is this working (?) Changed when moviepy 2.0.0
    video = video.with_effects([Crop(video, width = new_video_size[0], height = new_video_size[1], x_center = video_width / 2, y_center = video_height / 2)])

    # We resize it (up or down) to the desired 'size'
    video = video.resized((size[0], size[1]))

    if output_filename:
        video.write_videofile(output_filename)

    return video


from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import ImageClip
from typing import Union, Any


# TODO: I don't know if I should put this in another file or not...
# effect: segments.building.objects.effect.Effect = None
def generate_video_from_image(image: Union[str, Any], duration = 1, effect = None):
    """
    Genearates a ImageClip with the provided 'image_filename' that lasts the also
    provided 'duration'.

    This method will internally apply the 'effect' if provided. Check code or 
    developer guide to see available effects.
    """
    if not image:
        raise Exception('No "image" provided.')

    # if variable_is_type(image, str):
    #     # TODO: Check if image is actually an image
    #     # TODO: This image throws 'AttributeError: shape' moviepy error
    #     image = Image.open(image)
    # elif variable_is_type(image, np.ndarray):
    #     image = numpy_image_to_pil(image)

    if not PythonValidator.is_string(image):
        # TODO: Do something
        pass

    slide = ImageClip(image, duration = duration)
    slide = slide.with_fps(60)

    # TODO: This need work
    #slide = apply_effect(slide, effect)

    return slide


from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.color import Color, ColorString
from moviepy import Clip
from moviepy.video.fx import FadeIn as MoviepyFadeIn
from typing import Union


class FadeInEffect(Effect):
    """
    This effect will make the video appear progressively
    lasting the provided 'duration' time or the whole 
    clip duration if None 'duration' provided.

    The 'color' provided must be a valid color string, 
    array, tuple or Color instance, or will be set as 
    pure black if None provided.
    """
    def apply(self, video: Clip, duration: float, color: Union[list, tuple, str, Color] = None) -> Clip:
        # TODO: This is not working properly yet
        #PythonValidator.validate_method_params(BlinkEffect.apply, locals(), ['video'])
        video = VideoParser.to_moviepy(video)

        if not duration:
            duration = video.duration

        if color is None:
            color = Color.parse(ColorString.BLACK).as_rgb_array()
        else:
            color = Color.parse(color).as_rgb_array()

        return MoviepyFadeIn(duration, color).apply(video)


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_general_utils.image.region import Coordinate as BaseCoordinate
from yta_general_utils.programming.enum import YTAEnum as Enum
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import VideoClip


class CoordinateType(Enum):
    """
    Enum class to represent a coordinate type that 
    defines the way to make the calculations.
    """
    CORNER = 'corner'
    """
    The type of coordinate that represents the upper
    left corner of the object that uses the coordinate
    with this type.
    """
    CENTER = 'center'
    """
    The type of coordinate that represents the center of
    the object that uses the coordinate with this type.
    """

class Coordinate(BaseCoordinate):
    """
    Class to simplify the way to create a Coordinate
    instance by providing one of our pre-defined
    scene positions.
    """
    @staticmethod
    def from_position(position: 'Position'):
        """
        Initialize a VideoPosition with the given 'position' that
        will be computed according to our hypothetical default
        scenario of 1920x1080.
        """
        # Do not move import (cyclic import issue)
        from yta_multimedia.video.position import Position

        position: Position = Position.to_enum(position)

        position: VideoPosition = position.as_video_position(position)

        return BaseCoordinate(position.coordinate.x, position.coordinate.y)

class VideoPosition:
    """
    Coordinate to represent a the position in which we
    want to place a video. According to its 'type',
    we maybe want to put the center of the video in that
    position, or maybe the upper left corner.

    This is used to make easier the way we handle
    position parameters when trying to place a video
    in a specific place within a scene.

    TODO: Explain this better, please.
    """
    coordinate: Coordinate = None
    """
    The position (x, y) in which we want to position
    a video accoding to the 'type'.
    """
    type: CoordinateType = None
    """
    The type of the coordinate to be able to work
    properly with the 'x' and 'y' value and says
    what kind of representation is done by the
    coordinate.
    """

    @property
    def x(self):
        """
        Return the coordinate.x value.
        """
        return self.coordinate.x
    
    @property
    def y(self):
        """
        Return the coordinate.y value.
        """
        return self.coordinate.y

    def __init__(self, x: float, y: float, type: CoordinateType = CoordinateType.CORNER, is_normalized: bool = False):
        """
        Initialize a VideoPosition intance with raw values. If you
        have a Coordinate instance you can use the 
        'init_from_coordinate' method.
        """
        coordinate = Coordinate(x, y, is_normalized = is_normalized)

        self.coordinate = coordinate
        self.type = type

    def as_tuple(self):
        """
        Return the VideoPosition coordinate as a tuple (x, y).
        """
        return self.coordinate.as_tuple()
    
    def as_array(self):
        """
        Return the VideoPosition coordinate as an array [x, y].
        """
        return self.coordinate.as_array()
    
    @staticmethod
    def init_from_coordinate(coordinate: Coordinate, type: CoordinateType = CoordinateType.CORNER):
        """
        Initialize a VideoPosition instance by using a Coordinate
        instance.
        """
        if not PythonValidator.is_instance(coordinate, Coordinate):
            raise Exception('The provided "coordinate" parameter is not an instance of the Coordinate class.')
        
        type = CoordinateType.to_enum(type)

        return VideoPosition(coordinate.x, coordinate.y, type, coordinate.is_normalized)

    def get_position(self, video: VideoClip, background_video: VideoClip):
        """
        Translate the coordinate from our hypothetical
        1920x1080 scenario to the real one in which we
        are trying to place the provided 'video' in the
        also provided 'background_video' that could be
        different to our hypothetical scene.

        This method will return the coords (x, y) in
        which we need to place the 'video' to have its
        center in the desired ('x', 'y') position over
        the also provided 'background_video' by making
        some calculations as below.

        The coordinate is representing a specific 'x'
        and 'y' in a hypothetical scene of 1920x1080
        that could be pointing to the upper left corner
        or the center of a video according to the 'type'
        that has been set. With the provided 'video' and
        'background_video' we calculate the coord in 
        which the 'video' need to be positioned to
        actually fit the desired position as if it was
        the hypotethical 1920x1080 scenario.

        If we were representing the (100, 100) in that
        1920x1080 scenario and the new 'background_video'
        is 960x540, the new position would be (50, 50)
        because the scenario changed.
        """
        video = VideoParser.to_moviepy(video)
        background_video = VideoParser.to_moviepy(background_video)

        default_background_video = ClipGenerator.get_default_background_video()

        # Considering our hypothetical default 1920x1080 scene,
        # recalculate actual coords
        x = (int) (background_video.w * self.coordinate.x / default_background_video.w)
        y = (int) (background_video.h * self.coordinate.y / default_background_video.h)

        if self.type == CoordinateType.CENTER:
            # The video is pretended to be centered in that
            # position so we recalculate it to fit that
            # condition
            x -= int(video.w / 2)
            y -= int(video.h / 2)

        return (x, y)


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyResize
from yta_multimedia.video.edition.effect.moviepy.video_effect import MoviepyVideoEffect
from yta_general_utils.math.rate_functions import RateFunction
from moviepy import CompositeVideoClip, VideoFileClip, ImageClip, VideoClip, ColorClip
from typing import Union


# TODO: This can be replaced by using the MoviepyVideoEffect
# directly, but I keep this code by now to simplify the zoom
# use only
# TODO: Maybe remove in a future to force the use of the 
# general MoviepyVideoEffect
class ZoomVideoEffect(VideoEffect):
    """
    Creates a Zoom effect in the provided video.
    """
    @classmethod
    def apply(cls, video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], zoom_start: float, zoom_end: float, rate_func: type = RateFunction.linear):
        """
        Apply the effect on the provided 'video'.

        :param float zoom_start: The zoom at the start of the video, where a 1 is no zoom, a 0.8 is a zoom out of 20% and 1.1 is a zoom in of a 10%.

        :param float zoom_end: The zoom at the end of the video, where a 1 is no zoom, a 0.8 is a zoom out of 20% and 1.1 is a zoom in of a 10%.

        :param type rate_func: The rate function to apply in the animation effect. Must be one of the methods available in the RateFunction class.
        """
        return MoviepyVideoEffect.apply(video, MoviepyResize(zoom_start, zoom_end, rate_func))

    # TODO: 'apply_over_video' here (?)


from yta_multimedia.video.edition.effect.moviepy.position.objects.video_position import VideoPosition
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition
from yta_multimedia.video.edition.effect.m_effect import MEffect as Effect
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.edition.effect.moviepy.objects import validate_position
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.effect.moviepy.objects import MoviepyArgument, MoviepyWith
from yta_multimedia.audio.channels import synchronize_audio_pan_with_video_by_position
from yta_general_utils.math.rate_functions import RateFunction
from moviepy import CompositeVideoClip
from moviepy.Clip import Clip
from typing import Union


class MoveTestMoviepyEffect(Effect):
    """
    Effect of moving the element from one place in the screen (or outside
    of bounds) to another, linearly.
    """

    # TODO: This must inherit from an abstract class I think
    def apply(self, video: Clip, initial_position: Union[Position, VideoPosition], final_position: Union[Position, VideoPosition], arc_is_bottom: bool = False, max_arc_height: float = 300.0):
        """
        Applies the effect to the 'video' provided when initializing this
        effect class, and puts the video over a static black background
        image of 1920x1080.
        """
        return self.apply_over_video(video, ClipGenerator.get_default_background_video(duration = video.duration), initial_position, final_position, arc_is_bottom, max_arc_height)
    
    @classmethod
    def apply_over_video(cls, video: Clip, background_video: Clip, initial_position: Union[Position, VideoPosition, tuple] = Position.OUT_LEFT, final_position: Union[Position, VideoPosition, tuple] = Position.CENTER):
        video = VideoParser.to_moviepy(video)
        background_video = VideoParser.to_moviepy(background_video)

        validate_position(initial_position)
        validate_position(final_position)
        
        # We need the tuple position for the MoviepyArgument
        initial_position = get_moviepy_position(video, background_video, initial_position) 
        final_position = get_moviepy_position(video, background_video, final_position)

        arg = MoviepyArgument(initial_position, final_position, TFunctionSetPosition.arc, RateFunction.linear)
        video = MoviepyWith().apply_over_video(video, background_video, arg)

        # We pan the video audio by position
        video = synchronize_audio_pan_with_video_by_position(video.audio, video)

        return CompositeVideoClip([
            background_video,
            video
        ])


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.numpy_frame_helper import FrameMaskingMethod
from yta_multimedia.video.frames.video_frame import VideoFrame
from moviepy import VideoClip
from typing import Union


class MoviepyAlphaTransitionHandler:
    """
    Class to handle the functionality related to turning
    alpha clips into moviepy mask clips to be able to 
    apply them and create affects and transitions.
    """
    @staticmethod
    def alpha_clip_to_mask_clip(video: Union[str, VideoClip], masking_method: FrameMaskingMethod = FrameMaskingMethod.PURE_BLACK_AND_WHITE):
        """
        The provided alpha 'video' is turned into a moviepy
        mask clip that can be set in any other video by
        using the '.with_mask()' method.

        If you apply the resulting mask to a clip (called 
        'video2' in below) then you can do this below to
        enjoy an incredible transition, or using a black
        background instead of the 'video1' in below.

        CompositeVideoClip([
            video,
            video2
        ]).write_videofile('alpha_masked_transition.mp4')
        """
        video = VideoParser.to_moviepy(video, True, True)
        masking_method = FrameMaskingMethod.to_enum(masking_method)

        transparent_pixel_in_mask = False
        if video.mask:
            for alpha_frame in video.mask.iter_frames():
                if (alpha_frame < 1.0).any():
                    transparent_pixel_in_mask = True

        # If mask has transparency, we keep the mask
        mask_clip = video.mask
        if not transparent_pixel_in_mask:
            mask_clip_frames = [VideoFrame(frame).as_mask(masking_method) for frame in video.iter_frames()]

            mask_clip = VideoClip(lambda t: mask_clip_frames[int(t * video.fps)], is_mask = True).with_fps(video.fps)

        return mask_clip


from yta_multimedia.video.parser import VideoParser
from yta_general_utils.programming.parameter_validator import NumberValidator
from yta_general_utils.temp import create_temp_filename
from moviepy import VideoClip
from typing import Union


def crop_video_using_key_frame_second(video: Union[VideoClip, str], key_frame_second: float = None, duration: float = None, output_filename: Union[str, None] = None):
    video = VideoParser.to_moviepy(video)

    # TODO: Refactor and review this below
    if not NumberValidator.is_number_between(key_frame_second, 0, video.duration):
        # We use the middle of the video as the key frame
        key_frame_second = video.duration / 2
    if not duration:
        duration = video.duration
    if duration > video.duration or duration <= 0:
        duration = video.duration

    if duration < video.duration:
        start_second = 0
        end_second = video.duration
        # Only if we have to crop it already
        half_duration = duration / 2
        if key_frame_second - half_duration < 0:
            # Start in 0.0
            start_second = 0
            end_second = duration
        elif key_frame_second + half_duration > video.duration:
            # End in 'video.duration'
            start_second = video.duration - duration
            end_second = video.duration
        else:
            # Use 'key_frame_second' as center
            start_second = key_frame_second - half_duration
            end_second = key_frame_second + half_duration

        video = video.with_subclip(start_second, end_second)

    if output_filename:
        # TODO: Check extension
        tmp_audiofilename = create_temp_filename('temp-audio.m4a')
        # TODO: Do I really need all those parameters?
        video.write_videofile(
            output_filename,
            codec = "libx264",
            temp_audiofile = tmp_audiofilename,
            remove_temp = True,
            audio_codec = 'aac' # pcm_s16le or pcm_s32le
        )

    return video


# TODO: This package is 'position.utils.position' so it doesn't make sense
from yta_multimedia.video.edition.effect.moviepy.position.objects.video_position import VideoPosition
from yta_multimedia.video.position import Position
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import VideoFileClip, CompositeVideoClip, ImageClip, ColorClip
from typing import Union


def position_video_in(video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], background_video: Union[str, VideoFileClip, CompositeVideoClip, ImageClip, ColorClip], position: Union[Position, VideoPosition]):
    """
    Returns the 'video' positioned (with '.with_position(...)') to stay in 
    the provided 'position' without movement. It won't set any other
    property more than the duration (you will need to manually add
    '.with_duration()' or '.with_start()' if needed).

    This method will return the video positioned as a single element, so 
    make to wrap it properly in an array if it is part of a complex
    animation. 
    """
    if not video:
        raise Exception('No "video" provided.')
    
    if PythonValidator.is_string(video):
        if not FileValidator.file_is_video_file(video):
            raise Exception('Provided "video" is not a valid video file.')
        
        video = VideoFileClip(video)

    if not background_video:
        raise Exception('No "background_video" provided.')

    if PythonValidator.is_string(background_video):
        if not FileValidator.file_is_video_file(background_video):
            raise Exception('Provided "background_video" is not a valid video file.')
        
        background_video = VideoFileClip(background_video)

    if not PythonValidator.is_instance(position, Position):
        if not PythonValidator.is_instance(position, tuple) and len(position) != 2:
            raise Exception('Provided "position" is not a valid Position enum or (x, y) tuple.')
        
    position = get_moviepy_position(video, background_video, position)

    return video.with_position(position)


"""
    Coords related functions below
"""
def get_moviepy_position(video, background_video, position: Union[Position, VideoPosition, tuple]):
    """
    In the process of overlaying and moving the provided 'video' over
    the also provided 'background_video', this method calculates the
    (x, y) tuple position that would be, hypothetically, adapted from
    a 1920x1080 black color background static image. The provided 
    'position' will be transformed into the (x, y) tuple according
    to our own definitions in which the video (that starts in upper left
    corner) needs to be placed to fit the desired 'position'.
    """
    # TODO: Add 'video' and 'background_video' checkings
    if not video:
        raise Exception('No "video" provided.')
    
    if not background_video:
        raise Exception('No "background_video" provided.')
    
    if not position:
        raise Exception('No "position" provided.')
    
    if not PythonValidator.is_instance(position, [Position, VideoPosition, tuple]):
        raise Exception('Provided "position" is not a valid Position nor VideoPosition instance.')
    
    if PythonValidator.is_tuple(position) and len(position) != 2:
        # TODO: Maybe apply the normalization limits as limits
        # here for each position tuple element
        raise Exception('Provided "position" is a tuple but does not have 2 values.')
    
    position_tuple = position   # If tuple, here it is
    if PythonValidator.is_instance(position, Position):
        position_tuple = position.get_moviepy_position(video, background_video)
    elif PythonValidator.is_instance(position, VideoPosition):
        position_tuple = position.get_position(video, background_video)

    return position_tuple


from yta_multimedia.video.parser import VideoParser
from moviepy import vfx
from abc import ABC, abstractmethod


class VideoEffect(ABC):
    """
    Abstract class to be inherited by all my custom effects so I can 
    control they belong to this family.

    A video effect is an effect that is customly made by using 
    personal modifications, calculations, involving maybe some
    image manipulation, etc.

    A moviepy effect (or what I call like that) is an effect that is
    applied directly to the video by using only the moviepy editor
    and/or moviepy vfx module. It could be a simple moviepy effect
    made an object to simplify the work with it, or a more complex
    effect that is build with some different small effects.
    """

    @classmethod
    def parse_moviepy_video(cls, video, do_include_mask: bool = False):
        """
        Parses the provided video as a moviepy video clip returning it
        if valid or raising an Exception if invalid.
        """
        # TODO: This has to be removed to use the
        # VideoParser.to_moviepy directly
        return VideoParser.to_moviepy(video, do_include_mask = do_include_mask)
    
    @classmethod
    def get_moviepy_vfx_effect(cls, moviepy_effect_name: str):
        """
        Returns the moviepy vfx effect name corresponding to the provided
        'moviepy_effect_name'.
        """
        return getattr(vfx, moviepy_effect_name, None)

    @abstractmethod
    def apply(self):
        pass


from moviepy.Effect import Effect
from moviepy.Clip import Clip


class MEffect(Effect):
    """
    Effect to be applied on a single video without any
    other video dependence.

    This effect cannot be applied as moviepy effects as
    the structure is different.

    TODO: If I want to make my own moviepy effects that
    are able to be applied as moviepy does, I need to
    respect the structure and just manipulate the frames.
    So, those new custom moviepy effects sohuld be placed
    in other folder as this MEffect is my own effect that
    works in a different way.
    """
    # TODO: Is this working (?)
    result_must_replace: bool = False
    """
    This parameter indicates if this effect, when applied,
    must replace the original clip part or, if False, must
    be concatenated.
    """
    def apply_over_video(self, video: Clip, background_video: Clip):
        # TODO: Should I (?)
        pass


from yta_multimedia.video.parser import VideoParser
from yta_multimedia.video.frames.video_frame import VideoFrame
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import VideoClip, concatenate_videoclips
from typing import Union


class VideoHandler:
    """
    Class created to simplify and encapsulate the working process
    with moviepy videos.
    """
    @staticmethod
    def invert_video(video: Union[str, VideoClip]):
        """
        Invert the received 'video' (that must be a moviepy 
        mask or normal clip) and return it inverted as a
        VideoClip. If the provided 'video' is a mask, this 
        will be also a mask.

        If the 'clip' provided is a mask clip, remember to
        set it as the new mask of your main clip.

        This inversion is a process in which the numpy array
        values of each frame are inverted by substracting the
        highest value. If the frame is an RGB frame with 
        values between 0 and 255, it will be inverted by 
        doing 255 - X on each frame pixel value. If it is
        normalized and values are between 0 and 1 (it is a 
        mask clip frame), by doing 1 - X on each mask frame
        pixel value.
        """
        video = VideoParser.to_moviepy(video)

        mask_frames = [VideoFrame(frame).inverted() for frame in video.iter_frames()]

        return VideoClip(lambda t: mask_frames[int(t * video.fps)], is_mask = video.is_mask).with_fps(video.fps)
    

# TODO: Maybe this has been solved with the new
# moviepy 2.0.0 version
def concatenate_videos(videos: list[VideoClip]):
    """
    This is a fix for an issue with the moviepy 
    concatenate_videoclips method that seems to be
    failing with the concatenated video duration
    because of the floating point.

    Check this: 
    https://github.com/Zulko/moviepy/issues/646
    """
    if not PythonValidator.is_list(videos):
        videos = [videos]

    videos = [VideoParser.to_moviepy(video) for video in videos]

    # We do the moviepy concatenation
    video = concatenate_videoclips(videos)

    # We make a small substraction due to a float point
    # error when concatenating videos, to make sure we
    # use a value valid to be considered as the last
    # frame. The subclip method uses a (a, b] interval
    # to look for the frame, so as our .duration could
    # be a little greater than b, it would go for the
    # next interval and fail because of no frame 
    # available, we make sure we provide a value in the
    # range.
    video = video.with_subclip(0, video.duration - (1 / video.fps / 100000))

    return video


from yta_multimedia.video.edition.resize import fit_size
from yta_general_utils.file.checker import FileValidator
from yta_general_utils.file.enums import FileType
from yta_general_utils.file.filename import filename_is_type
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import VideoFileClip, CompositeVideoClip, ColorClip, ImageClip, ImageSequenceClip, VideoClip, AudioClip, concatenate_audioclips
from typing import Union


class VideoParser:
    """
    Class to simplify the way we parse video parameters.
    """
    @classmethod
    def to_moviepy(cls, video: Union[str, VideoFileClip, CompositeVideoClip, ColorClip, ImageClip], do_include_mask: bool = False, do_check_duration: bool = False, size: Union[tuple, None] = (1920, 1080)):
        """
        This method is a helper to turn the provided 'video' to a moviepy
        video type. If it is any of the moviepy video types specified in
        method declaration, it will be returned like that. If not, it will
        be load as a VideoFileClip if possible, or will raise an Exception
        if not.

        The 'do_include_mask' parameter includes the mask in the video if
        True value provided. The 'do_check_duration' parameter checks and
        updates the real video duration to fix a bug in moviepy lib.
        """
        if not video:
            raise Exception('No "video" provided.')
        
        # TODO: Maybe check if subclass of VideoClip
        if not PythonValidator.is_string(video) and not PythonValidator.is_instance(video, VideoFileClip) and not PythonValidator.is_instance(video, CompositeVideoClip) and not PythonValidator.is_instance(video, ColorClip) and not PythonValidator.is_instance(video, ImageClip) and not PythonValidator.is_instance(video, ImageSequenceClip) and not PythonValidator.is_instance(video, VideoClip):
            raise Exception('The "video" parameter provided is not a valid type. Check valid types in method declaration.')
        
        if PythonValidator.is_string(video):
            if not filename_is_type(video, FileType.VIDEO):
                raise Exception('The "video" parameter provided is not a valid video filename.')
            
            if not FileValidator.file_is_video_file(video):
                raise Exception('The "video" parameter is not a valid video file.')
            
            video = VideoFileClip(video, has_mask = do_include_mask)

        if do_check_duration:
            # We need to fix an existing bug in moviepy lib
            # see https://github.com/Zulko/moviepy/issues/1826
            video = verify_and_update_duration(video)

        # TODO: This below just adds a mask attribute but
        # without fps and empty, so it doesn't make sense
        # if do_include_mask and not video.mask:
        #     video = video.add_mask()

        if size:
            if not PythonValidator.is_tuple(size) or len(size) != 2:
                raise Exception(f'The provided "size" parameter "{str(size)}" is not a valid size.')
            
            # Adjust the strategy if needed
            video = fit_size(video, size)
        
        return video

# TODO: Maybe this has been solved with the new
# moviepy 2.0.0 version, I'm not sure but I will
# be using it not during the next weeks to check
# TODO: Very important method here below
def verify_and_update_duration(video: VideoClip):
    """
    Try to subclip the provided 'video' with the last
    frames detected by moviepy and check if the duration
    is the real one and updates it if not. This method
    returns the video updated to the its new duration.

    Moviepy has a bug in which some videos are detected
    larger than they actually are, and that makes the 
    software fails when trying to work with all those 
    wrong detected videos.

    When trying to subclip a moviepy video, the system
    will fail when the 't_start' parameter is larger or
    equal to the actual video duration (not the one
    detected by moviepy), so we are using that slow
    function to detect the real duration and updating
    the video accordingly.
    """
    frame_time = 1 / video.fps

    for frame_number in reversed(range(video.n_frames)):
        try:
            # Could I found the issue that the 'concatenate_videos'
            # method solves (?) Maybe I have to make a small
            # substraction, but here I am working with the 't_start'
            # not the 't_end'. I keep this comment here as an
            # advice. I hope it is not needed.
            real_duration = frame_number * frame_time
            video.with_subclip(real_duration, 9999)
            # 't_start' can't be the last frame but the penultimate
            # so the real duration is one 'frame_time' later
            real_duration += frame_time
            break
        except:
            pass

    video = video.with_subclip(0, real_duration)
    # Fix audio that can be problematic also
    video = video.with_audio(verify_and_update_audio_duration(video.audio))

    return video

# TODO: Move this to a better place
# TODO: This below is needed but it is apparently not
# working at all, because I still have the problem 
# when trying to combine two videos and the bugged 
# one (and fixed by this method) is one of them.
# Maybe new moviepy version fixes this (?)
def verify_and_update_audio_duration(audio: AudioClip):
    from yta_multimedia.audio.parser import AudioParser
    from yta_multimedia.audio.sound.generation.sound_generator import SoundGenerator

    audio = AudioParser.to_audiofileclip(audio)

    expected_duration = audio.duration

    cont = 0
    try:
        # TODO: Simplify this, it is too slow, maybe second
        # by second and later, when found the error, retry
        # frame by frame but in shorter range
        for _ in audio.iter_frames():
            cont += 1
    except:
        #print(f'Exception when {str(cont)} index.')
        actual_duration = cont / audio.fps
        rest_duration = expected_duration - actual_duration

        print(f'fixed audio with {str(rest_duration)} silence')
        
        audio = concatenate_audioclips([audio.with_subclip(0, cont), SoundGenerator.create_silence_audio(rest_duration)])

    return audio



from yta_multimedia.video.edition.effect.moviepy.t_function import TFunctionSetPosition, TFunctionResize, TFunctionRotate
from yta_multimedia.video.edition.effect.moviepy.position.objects.base_position_moviepy_effect import BasePositionMoviepyEffect
from yta_multimedia.video.edition.effect.moviepy.position.utils.position import get_moviepy_position
from yta_multimedia.video.edition.duration import set_video_duration
from yta_multimedia.video.edition.effect.moviepy.position.objects.video_position import VideoPosition
from yta_multimedia.video.edition.effect.moviepy.position import Position
from yta_multimedia.video.edition.effect.moviepy.mask import ClipGenerator
from yta_multimedia.video.parser import VideoParser
from yta_general_utils.math.rate_functions import RateFunction
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy.Clip import Clip
from typing import Union

import numpy as np
# TODO: Maybe unify all these to a MoviepyArgument (?) as
# they are very similar: 'start', 'end', 'rate_func'
# TODO: Maybe rename to MoviepyResize

class MoviepyResize:
    def __init__(self, initial_size: float, final_size: float, t_function: type = TFunctionResize.resize_from_to, rate_func: type = RateFunction.linear, *args, **kwargs):
        # TODO: Check that all params are provided and valid
        self.initial_size = initial_size
        self.final_size = final_size
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

class MoviepySetPosition:
    initial_position: VideoPosition = None
    final_position: VideoPosition = None

    def __init__(self, initial_position: tuple, final_position: float, t_function: type = TFunctionSetPosition.linear, rate_func: type = RateFunction.linear, *args, **kwargs):
        # TODO: Check that all params are provided and valid
        self.initial_position = initial_position
        self.final_position = final_position
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

    def apply(self, video):
        BasePositionMoviepyEffect.validate_position(self.initial_position)
        BasePositionMoviepyEffect.validate_position(self.final_position)

        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        video = video.with_position(lambda t: self.t_function(t, video.duration, get_moviepy_position(video, background_video, self.initial_position), get_moviepy_position(video, background_video, self.final_position), self.rate_func))

        return video
    
class MoviepyWithPosition:
    initial_position: VideoPosition = None
    final_position: VideoPosition = None

    def __init__(self, initial_position: tuple, final_position: float, t_function: type = TFunctionSetPosition.linear, rate_func: type = RateFunction.linear, *args, **kwargs):
        # TODO: Check that all params are provided and valid
        self.initial_position = initial_position
        self.final_position = final_position
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

    def apply(self, video: Clip):
        if not PythonValidator.is_instance(self.initial_position, [Position, VideoPosition]):
            raise Exception('Provided "position" is not a valid Position or VideoPosition.')
        
        if not PythonValidator.is_instance(self.final_position, [Position, VideoPosition]):
            raise Exception('Provided "position" is not a valid Position or VideoPosition.')
        
        background_video = ClipGenerator.get_default_background_video(duration = video.duration)

        frame_duration = video.duration / (video.duration * video.fps)
        ts = np.linspace(0, video.duration, int(video.duration * video.fps))

        sizes = [TFunctionResize.resize_from_to(t, video.duration, 0.1, 0.5, RateFunction.rush_into) for t in ts]
        positions = [TFunctionSetPosition.at_position(t, (500, 500), (video.w * sizes[i], video.h * sizes[i])) for i, t in enumerate(ts)]

        video = video.with_position(lambda t: self.t_function(t, video.duration, get_moviepy_position(video, background_video, self.initial_position), get_moviepy_position(video, background_video, self.final_position), self.rate_func))

        video = video.resized(lambda t: sizes[int(t // frame_duration)])
        video = video.with_position(lambda t: positions[int(t // frame_duration)])

        return video

class MoviepyRotate:
    # TODO: Check that all params are provided and valid
    def __init__(self, initial_rotation: int, final_rotation: int, t_function: type = TFunctionRotate.rotate_from_to, rate_func: type = RateFunction.linear, *args, **kwargs):
        self.initial_rotation = initial_rotation
        self.final_rotation = final_rotation
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: Set '*args' and '**kwargs'

# TODO: Keep this below and remove cod from above
class MoviepyArgument:
    """
    Arguments used for 'with_position', 'resized' and 'rotated'
    """
    def __init__(self, start: Union[int, tuple, float], end: Union[int, tuple, float], t_function: type, rate_func: RateFunction = RateFunction.linear):
        # TODO: 'start' and 'end' must be pure values which
        # means tuples if positions or numbers if resize or
        # rotate
        # TODO: Validate parameters
        self.start = start
        self.end = end
        self.t_function = t_function
        self.rate_func = rate_func
        # TODO: *args or **kwargs should be avoided as we must
        # define the 't_function' with the only parameters 
        # needed and make as much as we need. For example, we
        # should have '.arc_bottom' and '.arc_top' to avoid
        # passing a 'is_bottom' parameter because it would 
        # lost the common structure.

# TODO: Rename to MoviepyWithProcess or something more clear
class MoviepyWith:
    """
    Class to encapsulate the 'resized', 'with_position' and
    'rotated' functionality and to be able to make one depend
    on the other by pre-calculating the values and then using
    them when positioning, resizing and rotating.

    You should use the same RateFunction if you are planning
    to move and resize the clip at the same time.
    """
    @staticmethod
    def apply(video: Clip, with_position: Union[MoviepyArgument, None] = None, resized: Union[MoviepyArgument, None] = None, rotated: Union[MoviepyArgument, None] = None):
        """
        Apply the property effects on the provided 'video'. This
        method will use a default background video to apply the
        effect.

        Default background is 1920x1080.

        :param MoviepyArgument with_position: The position to apply (as a lambda t function).

        :param MoviepyArgument resized: The resize effect to apply (as a lambda t function).

        :param MoviepyArgument rotated: The rotation effect to apply (as a lambda t function).
        """
        return MoviepyWith.apply_over_video(video, ClipGenerator.get_default_background_video(duration = video.duration), with_position, resized, rotated)

    @staticmethod
    def apply_over_video(video: Clip, background_video: Clip, with_position: Union[MoviepyArgument, None] = None, resized: Union[MoviepyArgument, None] = None, rotated: Union[MoviepyArgument, None] = None):
        """
        Apply the property effects on the provided 'video'. This
        method will use the provided 'background_video' and 
        recalculate the position according to it.

        Background can be different to default scene of 1920x1080
        so positions will be adjusted to it.

        :param MoviepyArgument with_position: The position to apply (as a lambda t function).

        :param MoviepyArgument resized: The resize effect to apply (as a lambda t function).

        :param MoviepyArgument rotated: The rotation effect to apply (as a lambda t function).
        """
        video = VideoParser.to_moviepy(video, do_include_mask = True)
        background_video = VideoParser.to_moviepy(background_video, True)
        background_video = BasePositionMoviepyEffect.prepare_background_clip(background_video, video)

        # Prepare data for pre-calculations
        frame_duration = video.duration / (video.duration * video.fps)
        ts = np.linspace(0, video.duration, int(video.duration * video.fps))

        if resized:
            resizes = [resized.t_function(t, video.duration, resized.start, resized.end, resized.rate_func) for t in ts]
        else:
            # We build always this array because position calculation 
            # depends on it to be able to handle movement and resizing
            # at the same time
            resizes = [1 for _ in ts]

        positions = None
        if with_position:
            initial_position = get_moviepy_position(video, background_video, with_position.start)
            final_position = get_moviepy_position(video, background_video, with_position.end)
            positions = []
            for i, t in enumerate(ts):
                # Calculate each frame position based on each frame size
                initial_position = (initial_position[0] * resizes[i], initial_position[1] * resizes[i])
                final_position = (final_position[0] * resizes[i], final_position[1] * resizes[1])

                # TODO: This 'with_position.t_function' can have different
                # parameters to be able to work, so I don't know how to
                # dynamically handle this. For example, the 'at_position'
                # needs the video.size to be able to calculate where to
                # be centered...
                positions.append(with_position.t_function(t, video.duration, initial_position, final_position))
                
        rotations = None
        if rotated:
            rotations = [rotated.t_function(t, video.duration, rotated.start, rotated.end, rotated.rate_func) for t in ts]

        # I add frame_duration * 0.1 to make sure it fits the
        # next index
        video = video.resized(lambda t: resizes[int((t + frame_duration * 0.1) // frame_duration)])
        if positions:
            video = video.with_position(lambda t: positions[int((t + frame_duration * 0.1) // frame_duration)])
        if rotated:
            video = video.rotated(lambda t: rotations[int((t + frame_duration * 0.1) // frame_duration)])
        
        return video

# TODO: Maybe relocate these 2 methods from below
def validate_position(position: Union[Position, VideoPosition, tuple]):
    """
    Validates that the provided 'position' is an Position 
    or a VideoPosition, or raises an Exception if not.
    """
    if not PythonValidator.is_instance(position, [Position, VideoPosition, tuple]):
        raise Exception('Provided "position" is not a valid Position or VideoPosition.')
    
    if PythonValidator.is_tuple(position) and len(position) != 2:
        # TODO: Maybe apply the normalization limits as limits
        # here for each position tuple element
        raise Exception('Provided "position" is a tuple but does not have 2 values.')
    
def prepare_background_clip(background_video: Clip, video: Clip):
    """
    Prepares the provided 'background_video' by setting its duration to the same
    as the also provided 'video'. By default, the strategy is looping the 
    'background_video' if the 'video' duration is longer, or cropping it if it is
    shorter. This method returns the background_clip modified according to the
    provided 'video'.

    This method will raise an Exception if the provided 'video' or the provided
    'background_video' are not valid videos.
    """
    VideoParser.to_moviepy(background_video)
    VideoParser.to_moviepy(video)

    background_video = set_video_duration(background_video, video.duration)

    return background_video


from yta_multimedia.video.edition.effect.video_effect import VideoEffect
from yta_multimedia.video.edition.effect.moviepy.position.objects.video_position import VideoPosition
from yta_multimedia.video.position import Position
from yta_multimedia.video.edition.duration import set_video_duration
from yta_general_utils.programming.parameter_validator import PythonValidator
from moviepy import ColorClip, VideoFileClip, CompositeVideoClip, ImageClip, VideoClip
from typing import Union


class BasePositionMoviepyEffect(VideoEffect):
    """
    Class created to test position effects and building objects
    to simplify their use in our system.
    """

    @classmethod
    def validate_position(cls, position: Union[Position, VideoPosition]):
        """
        Validates that the provided 'position' is an Position 
        or a VideoPosition, or raises an Exception if not.
        """
        if not PythonValidator.is_instance(position, [Position, VideoPosition]):
            raise Exception('Provided "position" is not a valid Position or VideoPosition.')

    @classmethod
    def get_black_background_clip(cls, duration: float):
        # TODO: This should receive the main clip size
        return ColorClip((1920, 1080), [0, 0, 0], duration = duration)
    
    @classmethod
    def prepare_background_clip(cls, background_video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip], video: Union[str, VideoFileClip, VideoClip, CompositeVideoClip, ImageClip, ColorClip]):
        """
        Prepares the provided 'background_video' by setting its duration to the same
        as the also provided 'video'. By default, the strategy is looping the 
        'background_video' if the 'video' duration is longer, or cropping it if it is
        shorter. This method returns the background_clip modified according to the
        provided 'video'.

        This method will raise an Exception if the provided 'video' or the provided
        'background_video' are not valid videos.
        """
        VideoEffect.parse_moviepy_video(background_video)
        VideoEffect.parse_moviepy_video(video)

        background_video = set_video_duration(background_video, video.duration)

        return background_video

    @classmethod
    def apply(cls, video: Union[VideoFileClip, CompositeVideoClip, ImageClip]):
        """
        Applies the effect to the 'video' provided when initializing this
        effect class, and puts the video over a static black background
        image of 1920x1080.
        """
        pass

        #return BasePositionMoviepyEffect.apply_over_video(video, BasePositionMoviepyEffect.get_black_background_clip(video.duration))
    
    @classmethod
    def apply_over_video(cls, video: Union[VideoFileClip, CompositeVideoClip, ImageClip], background_clip: Union[VideoFileClip, CompositeVideoClip, ImageClip]):
        """
        This method parses the provided 'clip' and 'background_clip' to
        ensure they are valid moviepy clips and also sets the 
        'background_clip' duration to fit the provided 'clip' duration.
        """
        pass
        
        # VideoEffect.parse_moviepy_video(video)
        # VideoEffect.parse_moviepy_video(background_clip)

        # background_clip = set_video_duration(background_clip, video.duration)

        # return background_clip